import itertools
from enum import Enum
import numpy as np
import pandas as pd

import sys


##################
# params Enum
##################

class Optimizer(Enum):
    SGD = 1
    SGS_DECAY = 2
    ADAM = 3
    RMS_PROP = 4
    ADAMAX = 5
    NADAM = 5
    ADAGRAD = 5

class Process(Enum):
    PIN = 1
    APIN = 2

class Sampling(Enum):
    RANDOM = 1
    LATIN_HYPERCUBE = 2
    SPARSE_GRID = 3
    GRID_10 = 4
    GIRD_20 = 5
    GRID_LOCALP = 6
    GRID_ZERO = 7
    GRID_15 = 8
    GRID_20_LOCALP = 9
    GRID_20_ZERO = 10
    GRID_30 = 30
    GRID_50 = 12
    GRID_50_bis = 13

class Loss(Enum):
    MSE = 1
    MAE = 2

## add our models
class ParamsProcess:
    def draw_values(self, nb=1, smaller_range=False, mean_value=False):
        d = self.__dict__.copy()
        for k in d.keys():
            if smaller_range:
                min_ = d[k][0]
                max_ = d[k][1]
                delt = max_ - min_
                min_ = min_ + delt * 0.1
                max_ = max_ - delt * 0.1
            else:
                min_ = d[k][0]
                max_ = d[k][1]

            if (type(d[k][0]) == int) & (type(d[k][1]) == int):
                d[k] = np.random.randint(int(np.ceil(min_)), int(np.ceil(max_)) + 1, nb)
            else:
                if mean_value:
                    d[k] = np.array([(min_ + max_) / 2])
                else:
                    d[k] = np.random.uniform(min_, max_, nb)

        return d


class ParamsPin(ParamsProcess):
    def __init__(self):
        self.alpha = [0.0, 0.99]
        self.delta = [0.0, 0.99]
        self.epsilon_b = [100, 300]
        self.epsilon_s = [100, 300]
        self.mu = [100, 300]
        self.buy = [55, 700]
        self.sell = [55, 700]

class ParamsApin(ParamsProcess):
    def __init__(self) -> None:
        pass
    #In construction

class ParamsOption:
    def __init__(self):
        self.process = Process.PIN    

class ParamsModels:
    def __init__(self):
        self.save_dir = './model_save/'
        self.res_dir = './res/'

        self.name = Process.PIN
        self.normalize = True
        self.normalize_range = False
        # performance #
        # 8 layer: 0.8035
        # 7 layer: 0.78
        # 6 layer: 0.81 (VM)
        self.layers = [400,400,200,100] # 6 hidden layer
        #self.layers = [400,400,200,100] # 0.98
        self.batch_size = 256
        self.activation = "swish"
        self.opti = Optimizer.ADAM # use this
        self.loss = Loss.MSE

        self.learning_rate = 0.5e-3

        self.E = 15

class ParamsData:
    def __init__(self):
        self.path_sim_save = './data/'
        self.train_size = 3
        self.test_size = 10000
        # self.cross_vary_list = ["alpha","delta","epsilon_b","epsilon_s","mu","buy","sell"] # put only obsvervable state for backward compatibility
        self.cross_vary_list = ["buy","sell"]
        self.parallel = False

class Params:
    def __init__(self):
        self.name_detail = ''
        self.name = ''
        self.seed = 12345
        self.model = ParamsModels()
        self.data =  ParamsData()
        self.opt = ParamsOption()
  
        self.process = None
        self.update_process()
        # à remettre au prochain entrainement de model
        self.update_model_name()

    def update_process(self, process=None):
        if process is not None:
            self.opt.process = process

        if self.opt.process.name == Process.PIN.name:
            self.process = ParamsPin()

        if self.opt.process.name == Process.APIN.name:
            self.process = ParamsApin()


    def update_model_name(self):
        """
        change model name
        """
        n = self.name_detail
        n += 'Layer_'
        # for l in self.model.layers:
        #     n = n + str(l)+'_'
        if np.all(self.model.layers[0] == np.array(self.model.layers)):
            n = n + str(len(self.model.layers)) + 'L' + str(self.model.layers[0]) + '_'
        else:
            for l in self.model.layers:
                n = n + str(l) + '_'

        n += self.model.activation + '_Lr'
        n += str(self.model.learning_rate) + '_'
        # n += str(self.model.opti)+'_'

        n += self.model.opti.name + 'o' + self.model.loss.name + '_'
        n += 'BATCH_' + str(self.model.batch_size)

        n = n + 'tr_size_' + str(self.data.train_size) + 'CM'
        # for k in self.process.__dict__.keys():
        #     n = n + str(k) + str(self.process.__dict__[k][0]) + str(self.process.__dict__[k][1]) + '_'
        # n = n + 'tr_size_' + str(self.data.train_size) + '_te_size_' + str(self.data.test_size)

        n = n.replace('.', '_')
        n = n.replace('-', '_')
        self.name = n
        # In constrcution
        pass

    def print_values(self):
        """
        Print all parameters used in the model
        """

        for key, v in self.__dict__.items():
            try:
                print("#######",key,'#######')
                for key2, vv in v.__dict__.items():
                    print(key2, ":", vv)
            except:
                print(v)


    def save(self, save_dir, file_name="./parameters.p"):
       # simple save function that allows loading of deprecated parameters object
        df = pd.DataFrame(columns=['key', 'value'])

        for key, v in self.__dict__.items():
            try:
                for key2, vv in v.__dict__.items():
                    temp = pd.DataFrame(data=[str(key) + '_' + str(key2), vv], index=['key', 'value']).T
                    df = df.append(temp)

            except:
                temp = pd.DataFrame(data=[key, v], index=['key', 'value']).T
                df = df.append(temp)
        df.to_pickle("./"+save_dir + file_name,protocol=4)

    def load(self, load_dir, file_name="./parameters.p"):
        # simple load function that allows loading of deprecated parameters object
        df = pd.read_pickle(load_dir + file_name)
        # First check if this is an old pickle version, if so transform it into a df
        if type(df) != pd.DataFrame:
            loaded_par = df
            df = pd.DataFrame(columns=['key', 'value'])
            for key, v in loaded_par.__dict__.items():
                try:
                    for key2, vv in v.__dict__.items():
                        temp = pd.DataFrame(data=[str(key) + '_' + str(key2), vv], index=['key', 'value']).T
                        df = df.append(temp)

                except:
                    temp = pd.DataFrame(data=[key, v], index=['key', 'value']).T
                    df = df.append(temp)

        no_old_version_bug = True

        for key, v in self.__dict__.items():
            try:
                for key2, vv in v.__dict__.items():
                    t = df.loc[df['key'] == str(key) + '_' + str(key2), 'value']
                    if t.shape[0] == 1:
                        tt = t.values[0]
                        self.__dict__[key].__dict__[key2] = tt
                    else:
                        if no_old_version_bug:
                            no_old_version_bug = False
                            # print('#### Loaded parameters object is depreceated, default version will be used')
                        # print('Parameter', str(key) + '.' + str(key2), 'not found, using default: ',
                        #       self.__dict__[key].__dict__[key2])

            except:
                t = df.loc[df['key'] == str(key), 'value']
                if t.shape[0] == 1:
                    tt = t.values[0]
                    self.__dict__[key] = tt
                else:
                    if no_old_version_bug:
                        no_old_version_bug = False
                    #     print('#### Loaded parameters object is depreceated, default version will be used')
                    # print('Parameter', str(key), 'not found, using default: ', self.__dict__[key])

        self.update_process()

        # to ensure backward compatibility, we update her the cross_vary_list
        self.data.cross_vary_list = ["buy","sell"]

if __name__ == "__main__":
    params = Params()
    process = Process
    


# ML_modelpyhon -m venv 
# Created by Guillaume Pavé, at xx.xx.xx
import os
import pickle
import socket
from numpy import dtype
import tensorflow as tf
# remove tensorflow warnings
tf.get_logger().setLevel('ERROR')
tf.config.optimizer.set_jit(True)
tf.compat.v1.OptimizerOptions(cpu_global_jit=True)
tf.keras.mixed_precision.set_global_policy('mixed_float16')
print(f"Number of GPUs available: {tf.test.is_gpu_available()}")

from tensorflow import keras
from tensorflow.keras import layers
import tensorboard
import datetime
try:
    from parameters import *
except ModuleNotFoundError:
    from model.parameters import *
    
import pandas as pd
import numpy as np

tf.random.set_seed(12345)
class FirstLayer(tf.keras.layers.Layer):
    def __init__(self, num_outputs, par):
        super(FirstLayer, self).__init__(dtype='float64')
        self.num_outputs = num_outputs
        self.par = par

        c = []
        y = ['MLE']
        opt_data = self.par.data.cross_vary_list
        for cc in self.par.process.__dict__.keys():
            if (cc not in opt_data):
                c.append(cc)
        
        self.l1 = len(c)
        self.l2 = len(opt_data)

    def build(self,input_shape):
        self.kernel_par = self.add_weight("kernel_par",
                                          shape=[self.l1,
                                                 self.num_outputs], dtype=tf.float64)
        self.kernel_state = self.add_weight("kernel_data",
                                                shape=[self.l2,
                                                    self.num_outputs], dtype=tf.float64)

    def call(self,input):
        
        r = tf.matmul(input[0], self.kernel_par) + tf.matmul(tf.cast(input[1],tf.float64), self.kernel_state)
        
        r = tf.nn.swish(r)
        return r



class NetworkModel:
    def __init__(self, par: Params()):
        self.par = par
        self.model = None

        self.m = None
        self.std = None
        self.m_y = None
        self.std_y = None
        if socket.gethostname() in ['MBP-de-admin']:
            print("Guillaume's computer")

        self.save_dir = self.par.model.save_dir + '/' + self.par.name

    def normalize(self, X=None, y=None):
        if self.par.model.normalize:
            if X is not None:

                
                X = (X - self.m) / self.std

            if y is not None:
                pass

        return X, y

    def unnormalize(self, X=None, y=None):
        if self.par.model.normalize:
            if X is not None:
                if self.par.model.normalize_range:
                    X = X * self.x_range
                else:
                    X = (X * self.std) + self.m

            if y is not None:
                if type(y) == np.ndarray:
                    y = y * self.std_y.values + self.m_y.values + self
                else:
                    y = (y * self.std_y) + self.m_y
            
        return X, y

    # it works
    def split_state_data_par(self, df):

        opt_data = df[['buy', 'sell']]
        c = []
        for cc in self.par.process.__dict__.keys():
            # for cc in df.columns:
            if (cc not in opt_data.columns):
                c.append(cc)

        par_est = df[c]

        return [par_est, opt_data]

    def train(self): # In Construction
    #################
    # first get the data_plit col size
    #################

        c = []
        y = ['MLE'] # LE estimation
        opt_data = self.par.data.cross_vary_list

        for cc in self.par.process.__dict__.keys():
            if (cc not in opt_data):
                c.append(cc)

        c1 = len(c)
        c2 = len(c) + len(opt_data)

        d = self.par.process.__dict__
        
        m = {}
        std = {}

        for i, k in enumerate(d):
            m[k] = np.mean(d[k])
            std[k] = (((max(d[k]) - min(d[k])) ** 2) / 12) ** (1/2)
                
        self.m = pd.Series(m) #column in index
        self.std = pd.Series(std)
    
        ##################
        # prepare data sets
        ##################
        def tr(x):
            y = (x[:,:c1], x[:,c1:])
            
            return y

        print("=== Loading data ===")
        if self.par.opt.process.name == Process.PIN.name:
            data_dir = self.par.data.path_sim_save + 'PIN_MLE_new.txt'
        else:
            data_dir = self.par.data.path_sim_save + 'APIN_MLE.txt'
        
        data = pd.read_csv(data_dir,on_bad_lines='skip')
        print(f"shape of the data: {data.shape}")

        if self.model is None:
            self.create_nnet_model()


        # create splitting data
       
        y_data = data[y]
        x_data = data[c + opt_data]
        
        x_data, y_none = self.normalize(x_data)
        final_data = self.split_state_data_par(x_data)
        #Create a callback that saves the model's weights
        log_dir = "./model/logs/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1)
        cp_callback = tf.keras.callbacks.ModelCheckpoint(filepath=self.save_dir + '/', save_weights_only=True, verbose=0, save_best_only=True)
        print('start training for', self.par.model.E, 'epochs', flush=True)

        self.history_training = self.model.fit(x=final_data, y=y_data, validation_split=0.01,batch_size=self.par.model.batch_size, validation_batch_size=self.par.model.batch_size,epochs=self.par.model.E ,callbacks=[tensorboard_callback,cp_callback], verbose=1,use_multiprocessing=True)  # Pass callback to training
        

        self.history_training = pd.DataFrame(self.history_training.history)
        self.save()

    def predict(self, X):
        X, y = self.normalize(X, y=None)
        X = self.split_state_data_par(X)
        pred = self.model.predict(X)
        return pred

    def score(self,X,y):
        X, y = self.normalize(X, y)
        X = self.split_state_data_par(X)
        loss, mae, mse, r2 = self.model.evaluate(X, y, verbose=0)
        df = pd.DataFrame(np.array([mae,mse,r2]))
        df = df.T
        df.columns = ["mae","mse","r2"]
        return df

    def get_grad_and_mle(self, X, return_pred=False):
        """
        function that generate gradient and MLE
        """
        X, y = self.normalize(X, y=None)
        X = self.split_state_data_par(X)

        xx = [tf.convert_to_tensor(x.values) for x in X]
        
        with tf.GradientTape(persistent=True) as g:
            g.watch(xx[0])
            g.watch(xx[1])
            pred = self.model(xx)
            
        d = g.gradient(pred, xx[0])
        d1 = g.gradient(pred, xx[1])
        del g
        d = np.concatenate([d.numpy(), d1.numpy()], axis=1)
        col = list(X[0].columns) + list(X[1].columns)

        d = pd.DataFrame(d, columns=col, index=X[0].index)
        d = d*(1/self.std)

        if return_pred:
            return d, pred
        else:
            return d

    def save(self, other_save_dir=None):
        """
        function to save a file
        """
        self.par.save(save_dir=self.save_dir)

        with open(self.save_dir + '/m' + '.p', 'wb') as handle:
            pickle.dump(self.m, handle, protocol=pickle.HIGHEST_PROTOCOL)

        with open(self.save_dir + '/std' + '.p', 'wb') as handle:
            pickle.dump(self.std, handle, protocol=pickle.HIGHEST_PROTOCOL)

        with open(self.save_dir + '/m_y' + '.p', 'wb') as handle:
            pickle.dump(self.m_y, handle, protocol=pickle.HIGHEST_PROTOCOL)
        with open(self.save_dir + '/std_y' + '.p', 'wb') as handle:
            pickle.dump(self.std_y, handle, protocol=pickle.HIGHEST_PROTOCOL)
        self.history_training.to_pickle(self.save_dir + '/history.p')

    def load(self, n, other_save_dir=None):

        self.par.name = n
        if other_save_dir is None:
            temp_dir = self.par.model.save_dir + '' + self.par.name
        else:
            temp_dir = other_save_dir + '' + self.par.name

        if not os.path.exists(temp_dir):
            os.makedirs(temp_dir)

        par = Params()
        # par.load(load_dir=temp_dir)
        self.par = par

        with open(temp_dir + '/m' + '.p', 'rb') as handle:
            self.m = pickle.load(handle)

        with open(temp_dir + '/std' + '.p', 'rb') as handle:
            self.std = pickle.load(handle)

        with open(temp_dir + '/m_y' + '.p', 'rb') as handle:
            self.m_y = pickle.load(handle)
        with open(temp_dir + '/std_y' + '.p', 'rb') as handle:
            self.std_y = pickle.load(handle)

        self.history_training = pd.read_pickle(self.save_dir + '/history.p')

        if self.model is None:
            self.create_nnet_model()
        self.model.load_weights(self.save_dir + '/')


    def create_nnet_model(self):
        L = []
        for i, l in enumerate(self.par.model.layers):
            if i == 0:
                # L.append(tf.keras.layers.Dense(l, activation="swish", input_dim=2, input_shape=[len(self.par.process.__dict__)])) # trouver le moyen de changer ça
                L.append(FirstLayer(l,self.par))
            else:
                L.append(layers.Dense(l, activation= self.par.model.activation, dtype=tf.float64))
                L.append(tf.keras.layers.BatchNormalization())

            L.append(layers.Dense(1,dtype=tf.float64))
            self.model = keras.Sequential(L)

            # optimizer = tf.keras.optimizers.RMSprop(0.05)
        if self.par.model.opti == Optimizer.SGD:
            optimizer = tf.keras.optimizers.SGD(self.par.model.learning_rate)
        if self.par.model.opti == Optimizer.RMS_PROP:
            optimizer = tf.keras.optimizers.RMSprop(self.par.model.learning_rate)
        if self.par.model.opti == Optimizer.ADAM:
            optimizer = tf.keras.optimizers.Adam(self.par.model.learning_rate)
        if self.par.model.opti == Optimizer.NADAM:
            optimizer = tf.keras.optimizers.Nadam(self.par.model.learning_rate)
        if self.par.model.opti == Optimizer.ADAMAX:
            optimizer = tf.keras.optimizers.Adamax(self.par.model.learning_rate)
        if self.par.model.opti == Optimizer.ADAGRAD:
            optimizer = tf.keras.optimizers.Adamax(self.par.model.learning_rate)

        # optimizer = tf.keras.optimizers.Adam(0.00005/2)

        def r_square(y_true, y_pred):
            SS_res = tf.reduce_sum(tf.square(y_true - y_pred))
            SS_tot = tf.reduce_sum(tf.square(y_true - tf.reduce_mean(y_true)))
            return (1 - SS_res / (SS_tot + tf.keras.backend.epsilon()))

        if self.par.model.loss == Loss.MAE:
            self.model.compile(loss='mae', optimizer=optimizer, metrics=['mae', 'mse', r_square])
        if self.par.model.loss == Loss.MSE:
            self.model.compile(loss='mse', optimizer=optimizer, metrics=['mae', 'mse', r_square])
        

if __name__ == "__main__":
    par = Params()
    model = NetworkModel(par)
    model.train()


import setuptools
from distutils.core import setup

setup(
  name = 'DeepSurrogatepin',
  packages = ['model'],
  version = '0.4',
  license='MIT',
  setup_requires=['wheel'],
  description = 'Deep surrogate model for the probability of informed trading model',
  author = 'Guillaume Pavé',
  author_email = 'guillaumepave@gmail.com',
  url = 'https://github.com/GuillaumePv/pin_surrogate_model',
  # download_url = 'https://github.com/AntoineDidisheim/didipack/archive/v0.1.1.tar.gz',    # I explain this later on
  keywords = ['Machine learning, market microstructure'],
  install_requires=[            # I get to this in a second
          'numpy',
          'pandas',
          'tensorflow'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)


import os
from numpy import float64
import scipy.optimize as op
import pandas as pd
import time
from tqdm import tqdm
import tensorflow as tf
import tensorflow_probability as tfp
# https://www.tensorflow.org/probability?hl=fr
try:
    from ml_model import NetworkModel
except ModuleNotFoundError:
    from model.ml_model import NetworkModel

try:
    from parameters import *
except ModuleNotFoundError:
    from model.parameters import *



# add params args
class DeepSurrogate:
    def __init__(self):
        #print('Loading model', call_model_name)
        par_c = Params()
        print('par.model.save_dir')
        #par_c.load(self.par.model.save_dir + call_model_name)
        self.par_c = par_c
        self.c_model = NetworkModel(par_c)
        self.c_model.load(par_c.name)

        # à trouver le moyen de mettre dans une classe pour éviter les erreurs
        if self.par_c.opt.process.name == Process.PIN.name:
            data_dir = self.par_c.data.path_sim_save + 'PIN_MLE_new.txt'
        else:
            data_dir = self.par_c.data.path_sim_save + 'APIN_MLE.txt'

        # TO-DO: see to not load each time
        self.X = pd.read_csv(data_dir)
        self.Y = self.X['MLE']

        r = pd.Series(self.par_c.process.__dict__).apply(lambda x: (x[1]))
        m = self.c_model.m[r.index]
        s = self.c_model.std[r.index]
        pivot = (r - m) / s
        self.pivot = pivot[0]
        
        indexes = pd.Series(self.par_c.process.__dict__).index[:-2]
        
        self.means = self.c_model.m[indexes]
        self.std= self.c_model.std[indexes]
        
    def pre_process_X(self, X):
        # in construction
        X = X.copy()
        col = []
        for cc in self.par_c.process.__dict__.keys():
            col.append(cc)
        if X.shape[1] == len(col):
            X = pd.DataFrame(X, columns=col)
        else:
            assert False, 'wrong number of columns'
        
        return X
    
    # obtain gradients for each parameters
    def get_derivative(self, X):
        X = self.pre_process_X(X)
        return self.c_model.get_grad_and_mle(X)

    def get_model_score(self):
        print("=== score of the model ===")
        score = self.c_model.score(self.X.head(100000),self.Y.head(100000))
        print("=== compute score ====")
        score.to_latex("./results/table/result_model.tex",index=False)

    @tf.autograph.experimental.do_not_convert
    def get_pin(self,X):
        COL = ["alpha","delta","epsilon_b","epsilon_s","mu"]
        COL_B_S = self.par_c.data.cross_vary_list
        COL_PLUS = COL + COL_B_S
        init_x = self.means.values
        x = np.concatenate((init_x,X))
        df = pd.DataFrame([x],columns=COL_PLUS)
       
        def func_g(x_params):
            # génial pour faire sur plusieurs colonnes la même valeur
            df[COL] = x_params.numpy()
            # print("=== df ===")
            # print(df)
            grad, mle = self.c_model.get_grad_and_mle(df[COL_PLUS],True)
            # add possibilities to use it on several period
            # bug here
            loss_value = np.abs(np.sum(mle))
            g = grad.mean()[COL].values

            g = tf.convert_to_tensor(g)
            
            loss_value = tf.convert_to_tensor(loss_value)

            #print('---',loss_value,flush=True)
            return loss_value

       

        s = time.time()
        soln = tfp.optimizer.nelder_mead_minimize(objective_function=func_g, initial_vertex=init_x, max_iterations=50)
        soln_time = np.round((time.time() - s) / 60, 2)
        pred_par = soln.position.numpy()
        obj_value = soln.objective_value.numpy()
       
        # print(obj_value)
        # print(pred_par)
        # create a table with PIN, buys and sells
        PIN = np.round((pred_par[0]*pred_par[4])/((pred_par[0]*pred_par[4])+pred_par[2]+pred_par[3]),4)
        return PIN
    
if __name__ == '__main__':
    deepsurrogate = DeepSurrogate()
    deepsurrogate.get_model_score()


        
